'''Client to interact with AirNow Air Quality API'''
from json.decoder import JSONDecodeError
from typing import Optional, cast

from aiohttp import ClientSession, ClientTimeout

from .errors import AirNowError, EmptyResponseError, InvalidJsonError, \
    InvalidKeyError
from .forecast import Forecast
from .observation import Observations

API_BASE_URL: str = 'https://www.airnowapi.org'
API_DEFAULT_TIMEOUT: int = 10


class WebServiceAPI:
    '''Client to interact with AirNow API'''
    def __init__(
        self, api_key: str, *, session: Optional[ClientSession] = None
    ) -> None:
        '''Initialize with Client Session and API Key'''
        self._api_key: Optional[str] = api_key
        self._session: Optional[ClientSession] = session

        self.forecast = Forecast(self._get)
        self.observations = Observations(self._get)

    async def _get(
        self, endpoint: str, *, base_url: str = API_BASE_URL, **kwargs
    ) -> list:
        '''Run a Request against the API'''
        kwargs.setdefault('params', {})
        kwargs['params']['API_KEY'] = self._api_key
        kwargs['params']['format'] = 'application/json'

        use_running_session = self._session and not self._session.closed

        session: ClientSession
        if use_running_session:
            session = cast(ClientSession, self._session)
        else:
            session = ClientSession(
                timeout=ClientTimeout(total=API_DEFAULT_TIMEOUT)
            )

        try:
            async with session.get(
                f'{base_url}/{endpoint}', **kwargs
            ) as resp:
                data = await resp.json(content_type=None)

        except JSONDecodeError:
            # The response can't be parsed as JSON, so we'll use its body text
            # in an error:
            response_text = await resp.text()
            raise InvalidJsonError(response_text)

        finally:
            if not use_running_session:
                await session.close()

        if isinstance(data, dict) and 'WebServiceError' in data:
            # Process an Error Message from the API server
            wsErr = data['WebServiceError']
            if isinstance(wsErr, list) and len(wsErr) > 0:
                if 'Message' in wsErr[0]:
                    if wsErr[0]['Message'] == 'Invalid API key':
                        raise InvalidKeyError(wsErr[0]['Message'])
                    elif wsErr[0]['Message'] == 'Request not authenticated':
                        raise InvalidKeyError(wsErr[0]['Message'])
                    else:
                        raise AirNowError(wsErr[0]['Message'])
                else:
                    raise AirNowError(str(wsErr[0]))
            else:
                raise AirNowError(str(wsErr))

        elif not isinstance(data, list):
            # We should get a list of Observation or Forecast objects
            raise InvalidJsonError(
                'Unexpected response type: %s' % (type(data))
            )

        elif len(data) == 0:
            # No objects were returned
            raise EmptyResponseError('No data was returned')

        # Return list of objects for further processing
        return data



'''pyAirNow pytest Fixtures'''
import pytest
import re

from aioresponses import aioresponses

from .mock_api import mock_airnow_api


@pytest.fixture
def mock_aioresponse():
    with aioresponses() as m:
        yield m


@pytest.fixture
def mock_airnowapi(mock_aioresponse):
    url_pattern = re.compile(r'^https://www\.airnowapi\.org/(.*)$')
    mock_aioresponse.get(url_pattern, callback=mock_airnow_api)



import pytest

from aiohttp import ClientSession
from pyairnow import WebServiceAPI

from .mock_api import MOCK_API_KEY


@pytest.mark.asyncio
async def test_api(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.zipCode('90001')

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'forecast'
    assert data[0]['mode'] == 'zipCode'
    assert data[0]['when'] is None

    assert 'zipCode' in data[0]['query']
    assert data[0]['query']['zipCode'] == '90001'


@pytest.mark.asyncio
async def test_api_with_session(mock_airnowapi):
    session = ClientSession()
    client = WebServiceAPI(MOCK_API_KEY, session=session)
    data = await client.forecast.zipCode(90001)

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'forecast'
    assert data[0]['mode'] == 'zipCode'
    assert data[0]['when'] is None

    assert 'zipCode' in data[0]['query']
    assert data[0]['query']['zipCode'] == '90001'



import pytest

from pyairnow import WebServiceAPI
from pyairnow.errors import AirNowError, EmptyResponseError, \
    InvalidJsonError, InvalidKeyError

from .mock_api import MOCK_API_KEY


@pytest.mark.asyncio
async def test_api_not_authenticated(mock_airnowapi):
    client = WebServiceAPI('')
    with pytest.raises(InvalidKeyError) as exc_info:
        await client.forecast.zipCode(90001)
        assert 'not authenticated' in str(exc_info.value)


@pytest.mark.asyncio
async def test_api_invalid_key(mock_airnowapi):
    client = WebServiceAPI('123ABC')
    with pytest.raises(InvalidKeyError) as exc_info:
        await client.forecast.zipCode(90001)
        assert 'Invalid API key' in str(exc_info.value)


@pytest.mark.asyncio
async def test_api_bad_json(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    with pytest.raises(InvalidJsonError):
        await client.forecast.zipCode('bad_json')


@pytest.mark.asyncio
async def test_api_bad_json_2(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    with pytest.raises(InvalidJsonError):
        await client.forecast.zipCode('dict')


@pytest.mark.asyncio
async def test_api_empty_response(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    with pytest.raises(EmptyResponseError):
        await client.forecast.zipCode('empty')


@pytest.mark.asyncio
async def test_api_unexpected_error(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    with pytest.raises(AirNowError) as exc_info:
        await client.forecast.zipCode('error')
        assert 'Client Error' in str(exc_info.value)


@pytest.mark.asyncio
async def test_api_unexpected_error_1(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    with pytest.raises(AirNowError) as exc_info:
        await client.forecast.zipCode('error1')
        assert 'Internal Server Error' in str(exc_info.value)


@pytest.mark.asyncio
async def test_api_unexpected_error_2(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    with pytest.raises(AirNowError) as exc_info:
        await client.forecast.zipCode('error2')
        assert 'Internal Server Error' in str(exc_info.value)



import datetime
import pytest

from pyairnow import WebServiceAPI

from .mock_api import MOCK_API_KEY


@pytest.mark.asyncio
async def test_api_forecast_zipcode(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.zipCode(90001)

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'forecast'
    assert data[0]['mode'] == 'zipCode'
    assert data[0]['when'] is None

    assert 'zipCode' in data[0]['query']
    assert data[0]['query']['zipCode'] == '90001'


@pytest.mark.asyncio
async def test_api_forecast_zipcode_distance(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.zipCode(90001, distance=100)

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'forecast'
    assert data[0]['mode'] == 'zipCode'
    assert data[0]['when'] is None

    assert 'zipCode' in data[0]['query']
    assert data[0]['query']['zipCode'] == '90001'

    assert 'distance' in data[0]['query']
    assert data[0]['query']['distance'] == '100'


@pytest.mark.asyncio
async def test_api_forecast_zipcode_date_str(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.zipCode(90001, date='2020-09-01')

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'forecast'
    assert data[0]['mode'] == 'zipCode'
    assert data[0]['when'] is None

    assert 'zipCode' in data[0]['query']
    assert data[0]['query']['zipCode'] == '90001'

    assert 'date' in data[0]['query']
    assert data[0]['query']['date'] == '2020-09-01'


@pytest.mark.asyncio
async def test_api_forecast_zipcode_date_date(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.zipCode(
        90001,
        date=datetime.date(2020, 9, 1)
    )

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'forecast'
    assert data[0]['mode'] == 'zipCode'
    assert data[0]['when'] is None

    assert 'zipCode' in data[0]['query']
    assert data[0]['query']['zipCode'] == '90001'

    assert 'date' in data[0]['query']
    assert data[0]['query']['date'] == '2020-09-01'


@pytest.mark.asyncio
async def test_api_forecast_zipcode_date_datetime(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.zipCode(
        90001,
        date=datetime.datetime(2020, 9, 1, 11, 45, 1)
    )

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'forecast'
    assert data[0]['mode'] == 'zipCode'
    assert data[0]['when'] is None

    assert 'zipCode' in data[0]['query']
    assert data[0]['query']['zipCode'] == '90001'

    assert 'date' in data[0]['query']
    assert data[0]['query']['date'] == '2020-09-01'


@pytest.mark.asyncio
async def test_api_forecast_ll(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.latLong(34.053718, -118.244842)

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'forecast'
    assert data[0]['mode'] == 'latLong'
    assert data[0]['when'] is None

    assert 'latitude' in data[0]['query']
    assert data[0]['query']['latitude'] == '34.053718'
    assert 'longitude' in data[0]['query']
    assert data[0]['query']['longitude'] == '-118.244842'


@pytest.mark.asyncio
async def test_api_forecast_ll_distance(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.latLong(
        34.053718, -118.244842,
        distance=120
    )

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'forecast'
    assert data[0]['mode'] == 'latLong'
    assert data[0]['when'] is None

    assert 'latitude' in data[0]['query']
    assert data[0]['query']['latitude'] == '34.053718'
    assert 'longitude' in data[0]['query']
    assert data[0]['query']['longitude'] == '-118.244842'

    assert 'distance' in data[0]['query']
    assert data[0]['query']['distance'] == '120'


@pytest.mark.asyncio
async def test_api_forecast_ll_date_str(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.latLong(
        34.053718, -118.244842,
        date='2020-09-01'
    )

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'forecast'
    assert data[0]['mode'] == 'latLong'
    assert data[0]['when'] is None

    assert 'latitude' in data[0]['query']
    assert data[0]['query']['latitude'] == '34.053718'
    assert 'longitude' in data[0]['query']
    assert data[0]['query']['longitude'] == '-118.244842'

    assert 'date' in data[0]['query']
    assert data[0]['query']['date'] == '2020-09-01'


@pytest.mark.asyncio
async def test_api_forecast_ll_date_date(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.latLong(
        34.053718, -118.244842,
        date=datetime.date(2020, 9, 1)
    )

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'forecast'
    assert data[0]['mode'] == 'latLong'
    assert data[0]['when'] is None

    assert 'latitude' in data[0]['query']
    assert data[0]['query']['latitude'] == '34.053718'
    assert 'longitude' in data[0]['query']
    assert data[0]['query']['longitude'] == '-118.244842'

    assert 'date' in data[0]['query']
    assert data[0]['query']['date'] == '2020-09-01'


@pytest.mark.asyncio
async def test_api_forecast_ll_date_datetime(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.forecast.latLong(
        34.053718, -118.244842,
        date=datetime.datetime(2020, 9, 1, 11, 45, 0)
    )

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'forecast'
    assert data[0]['mode'] == 'latLong'
    assert data[0]['when'] is None

    assert 'latitude' in data[0]['query']
    assert data[0]['query']['latitude'] == '34.053718'
    assert 'longitude' in data[0]['query']
    assert data[0]['query']['longitude'] == '-118.244842'

    assert 'date' in data[0]['query']
    assert data[0]['query']['date'] == '2020-09-01'



'''Mocked AirNow API'''
import re

from aioresponses import CallbackResult

MOCK_API_KEY = '01234567-89AB-CDEF-0123-456789ABCDEF'
RE_ENDPOINTS = re.compile(
    '^/?aq/(forecast|observation)/(zipCode|latLong)(/(current))?/?$'
)


def mock_airnow_api(url, **kwargs):
    '''
    Mock the AirNow API
    '''
    m = RE_ENDPOINTS.match(url.path)
    if m is None:
        return CallbackResult(status=404)

    # Parse URL path
    ep_type = m.group(1)
    ep_mode = m.group(2)
    ep_when = m.group(4)

    # Check API Key
    if 'API_KEY' not in url.query:
        return CallbackResult(status=401, payload=dict(
            WebServiceError=[
                dict(
                    Message='Request not authenticated'
                ),
            ],
        ))

    if url.query['API_KEY'] != MOCK_API_KEY:
        return CallbackResult(status=401, payload=dict(
            WebServiceError=[
                dict(
                    Message='Invalid API key'
                ),
            ],
        ))

    # A zip code with value "empty" will return an empty list
    if 'zipCode' in url.query and url.query['zipCode'] == 'empty':
        return CallbackResult(payload=[])

    # A zip code with value "bad_json" will return invalid json
    if 'zipCode' in url.query and url.query['zipCode'] == 'bad_json':
        return CallbackResult(body='Bad JSON Test')

    # A zip code with value "error" will return a JSON error message
    if 'zipCode' in url.query and url.query['zipCode'] == 'error':
        return CallbackResult(status=400, payload=dict(
            WebServiceError=[
                dict(
                    Message='Client Error'
                ),
            ],
        ))

    # A zip code with value "error1" will return a JSON error message
    if 'zipCode' in url.query and url.query['zipCode'] == 'error1':
        return CallbackResult(status=400, payload=dict(
            WebServiceError=[
                'Client Error'
            ],
        ))

    # A zip code with value "error2" will return a non-structured error message
    if 'zipCode' in url.query and url.query['zipCode'] == 'error2':
        return CallbackResult(status=500, payload=dict(
            WebServiceError='Internal Server Error'
        ))

    # A zip code with value "dict" will return a JSON dictionary instead of
    # an array
    if 'zipCode' in url.query and url.query['zipCode'] == 'dict':
        return CallbackResult(payload=dict(status='OK'))

    # Return JSON with the endpoint and query information
    payload = dict(
        type=ep_type,
        mode=ep_mode,
        when=ep_when,
        query=dict(url.query),
    )
    return CallbackResult(payload=[payload])



import pytest

from pyairnow.conv import aqi_to_concentration, concentration_to_aqi


def test_convert_pm25():
    assert aqi_to_concentration(0, 'PM2.5') == 0
    assert aqi_to_concentration(25, 'PM2.5') == 4.5
    assert aqi_to_concentration(50, 'PM2.5') == 9.0
    assert aqi_to_concentration(51, 'PM2.5') == 9.1
    assert aqi_to_concentration(86, 'PM2.5') == 27.9
    assert aqi_to_concentration(144, 'PM2.5') == 53.0
    assert aqi_to_concentration(500, 'PM2.5') == 325.4
    assert aqi_to_concentration(501, 'PM2.5') == 325.9

    assert concentration_to_aqi(0, 'PM2.5') == 0
    assert concentration_to_aqi(9.0, 'PM2.5') == 50
    assert concentration_to_aqi(9.1, 'PM2.5') == 51
    assert concentration_to_aqi(27.9, 'PM2.5') == 86
    assert concentration_to_aqi(53.0, 'PM2.5') == 144
    assert concentration_to_aqi(325.4, 'PM2.5') == 500
    assert concentration_to_aqi(325.9, 'PM2.5') == 501


def test_convert_invalid_negative_aqi():
    with pytest.raises(ValueError):
        aqi_to_concentration(-1, 'PM2.5')


def test_convert_invalid_negative_o3():
    with pytest.raises(ValueError):
        concentration_to_aqi(-0.0008, 'O3')



'''Retrieve a list of Current Observations'''
from typing import Callable, Coroutine, Optional, Union


class Observations:
    '''
    Class to retrieve the current air quality observations by zip code or by
    latitude and longitude.
    '''
    def __init__(self, request: Callable[..., Coroutine]) -> None:
        self._request = request

    async def zipCode(
        self,
        zipCode: str,
        *,
        distance: Optional[int] = None
    ) -> list:
        '''Request current observation for zip code'''
        params: dict = dict(zipCode=zipCode)
        if distance:
            params['distance'] = distance

        return await self._request(
            'aq/observation/zipCode/current',
            params=params
        )

    async def latLong(
        self,
        latitude: Optional[Union[float, str]] = None,
        longitude: Optional[Union[float, str]] = None,
        *,
        distance: Optional[int] = None,
    ) -> list:
        '''Request current observation for latitude/longitude'''
        params: dict = dict(
            latitude=str(latitude),
            longitude=str(longitude),
        )
        if distance:
            params['distance'] = distance

        return await self._request(
            'aq/observation/latLong/current',
            params=params
        )



'''Convert AQI to and from Pollutant Concentrations'''
from typing import TypedDict


class ConcBreakpointPair(TypedDict):
    bL: float
    bH: float


class EPATableRow(TypedDict):
    iL: int
    iH: int
    breakpoints: dict[str, ConcBreakpointPair]


EPA_TABLE: list[EPATableRow] = [
    {
        'iL': 0,
        'iH': 50,
        'breakpoints': {
            'O3':       { 'bL': 0.000,    'bH': 0.054   },
            'PM2.5':    { 'bL': 0.0,      'bH': 9.0     },
            'PM10':     { 'bL': 0,        'bH': 54      },
            'CO':       { 'bL': 0.0,      'bH': 4.4     },
            'SO2':      { 'bL': 0,        'bH': 35      },
            'NO2':      { 'bL': 0,        'bH': 53      },
        },
    },
    {
        'iL': 51,
        'iH': 100,
        'breakpoints': {
            'O3':       { 'bL': 0.055,    'bH': 0.070   },
            'PM2.5':    { 'bL': 9.1,      'bH': 35.4    },
            'PM10':     { 'bL': 55,       'bH': 154     },
            'CO':       { 'bL': 4.5,      'bH': 9.4     },
            'SO2':      { 'bL': 36,       'bH': 75      },
            'NO2':      { 'bL': 54,       'bH': 100     },
        },
    },
    {
        'iL': 101,
        'iH': 150,
        'breakpoints': {
            'O3':       { 'bL': 0.071,    'bH': 0.085   },
            'PM2.5':    { 'bL': 35.5,     'bH': 55.4    },
            'PM10':     { 'bL': 155,      'bH': 254     },
            'CO':       { 'bL': 9.5,      'bH': 12.4    },
            'SO2':      { 'bL': 76,       'bH': 185     },
            'NO2':      { 'bL': 101,      'bH': 360     },
        },
    },
    {
        'iL': 151,
        'iH': 200,
        'breakpoints': {
            'O3':       { 'bL': 0.086,    'bH': 0.105   },
            'PM2.5':    { 'bL': 55.5,     'bH': 125.4   },
            'PM10':     { 'bL': 255,      'bH': 354     },
            'CO':       { 'bL': 12.5,     'bH': 15.4    },
            'SO2':      { 'bL': 186,      'bH': 304     },
            'NO2':      { 'bL': 361,      'bH': 649     },
        },
    },
    {
        'iL': 201,
        'iH': 300,
        'breakpoints': {
            'O3':       { 'bL': 0.106,    'bH': 0.200   },
            'PM2.5':    { 'bL': 125.5,    'bH': 225.4   },
            'PM10':     { 'bL': 355,      'bH': 424     },
            'CO':       { 'bL': 15.5,     'bH': 30.4    },
            'SO2':      { 'bL': 305,      'bH': 604     },
            'NO2':      { 'bL': 650,      'bH': 1249    },
        },
    },
    {
        'iL': 301,
        'iH': 500,
        'breakpoints': {
            # Ozone here changes from 8h to 1h reporting
            'O3':       { 'bL': 0.201,    'bH': 0.604   },
            'PM2.5':    { 'bL': 225.5,    'bH': 325.4   },
            'PM10':     { 'bL': 425,      'bH': 604     },
            'CO':       { 'bL': 30.5,     'bH': 50.4    },
            'SO2':      { 'bL': 605,      'bH': 1004    },
            'NO2':      { 'bL': 1250,     'bH': 2049    },
        },
    },
]

EPA_ROUND_FNS = {
    'O3':       lambda c: round(c, 3),
    'PM2.5':    lambda c: round(c, 1),
    'PM10':     lambda c: int(round(c, 0)),
    'CO':       lambda c: round(c, 1),
    'SO2':      lambda c: int(round(c, 0)),
    'NO2':      lambda c: int(round(c, 0)),
}


def aqi_to_concentration(aqi: int, pollutant: str) -> float:
    '''Convert AQI (0-500) to Pollutant Concentration'''
    if aqi < 0:
        raise ValueError('AQI must be greater than 0')

    max_breakpoint_aqi = EPA_TABLE[-1]['iH']
    for row in EPA_TABLE:
        iL = row['iL']
        iH = row['iH']

        # The min() comparison here selects the last breakpoint row when
        # AQI > 500. See the comment in `concentration_to_aqi`.
        if iL <= aqi and iH >= min(aqi, max_breakpoint_aqi):
            bL = row['breakpoints'][pollutant]['bL']
            bH = row['breakpoints'][pollutant]['bH']
            cp = float(aqi - iL) * (bH - bL) / (iH - iL) + bL
            return EPA_ROUND_FNS[pollutant](cp)

    raise RuntimeError('No matching row found in EPA_TABLE')


def concentration_to_aqi(conc: float, pollutant: str) -> int:
    '''Convert Pollutant Concentration to AQI'''
    if conc < 0:
        raise ValueError('Concentration must be positive')

    rounded = EPA_ROUND_FNS[pollutant](conc)
    max_breakpoint_pollutant_conc = (
        EPA_TABLE[-1]['breakpoints'][pollutant]['bH']
    )
    for row in EPA_TABLE:
        bL = row['breakpoints'][pollutant]['bL']
        bH = row['breakpoints'][pollutant]['bH']
        # As per EPA:
        #   AQI values greater than 500 should be calculated using
        #   [this algorithm] and the concentration specified for the
        #   AQI value of 500.
        # The min() comparison here selects the last breakpoint row
        # in those cases.
        if bL <= rounded and bH >= min(rounded, max_breakpoint_pollutant_conc):
            iL = row['iL']
            iH = row['iH']

            return int(round(float(iH - iL) / (bH - bL) * (conc - bL) + iL, 0))

    raise ValueError(
        f'Invalid concentration {rounded} for pollutant {pollutant}'
    )



import pytest

from pyairnow import WebServiceAPI

from .mock_api import MOCK_API_KEY


@pytest.mark.asyncio
async def test_api_observations_zipcode(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.observations.zipCode(90001)

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'observation'
    assert data[0]['mode'] == 'zipCode'
    assert data[0]['when'] == 'current'

    assert 'zipCode' in data[0]['query']
    assert data[0]['query']['zipCode'] == '90001'


@pytest.mark.asyncio
async def test_api_observations_zipcode_distance(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.observations.zipCode(90001, distance=100)

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'observation'
    assert data[0]['mode'] == 'zipCode'
    assert data[0]['when'] == 'current'

    assert 'zipCode' in data[0]['query']
    assert data[0]['query']['zipCode'] == '90001'

    assert 'distance' in data[0]['query']
    assert data[0]['query']['distance'] == '100'


@pytest.mark.asyncio
async def test_api_observations_ll(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.observations.latLong(34.053718, -118.244842)

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'observation'
    assert data[0]['mode'] == 'latLong'
    assert data[0]['when'] == 'current'

    assert 'latitude' in data[0]['query']
    assert data[0]['query']['latitude'] == '34.053718'
    assert 'longitude' in data[0]['query']
    assert data[0]['query']['longitude'] == '-118.244842'


@pytest.mark.asyncio
async def test_api_observations_ll_distance(mock_airnowapi):
    client = WebServiceAPI(MOCK_API_KEY)
    data = await client.observations.latLong(
        34.053718, -118.244842,
        distance=120
    )

    assert isinstance(data, list)
    assert len(data) == 1

    assert data[0]['type'] == 'observation'
    assert data[0]['mode'] == 'latLong'
    assert data[0]['when'] == 'current'

    assert 'latitude' in data[0]['query']
    assert data[0]['query']['latitude'] == '34.053718'
    assert 'longitude' in data[0]['query']
    assert data[0]['query']['longitude'] == '-118.244842'

    assert 'distance' in data[0]['query']
    assert data[0]['query']['distance'] == '120'



'''Test Script/CLI Interface to pyairnow'''
import argparse
import asyncio

from pyairnow import WebServiceAPI


def optionalArgs(args):
    '''Load optional keyword arguments (date and dict)'''
    kwargs = dict()
    if args.date is not None:
        kwargs['date'] = args.date
    if args.distance is not None:
        kwargs['distance'] = args.distance

    return kwargs


def llArgs(args):
    '''Load latitude and longitude from the location argument'''
    try:
        lat, long = args.split(',')
    except ValueError:
        raise ValueError(
            'Invalid location string, expected <latitude>,<longitude>'
        )

    return [float(lat.strip()), float(long.strip())]


async def get_forecast_zip(args: argparse.Namespace) -> None:
    '''Print forecast information for a zip code'''
    client = WebServiceAPI(args.api_key)
    data = await client.forecast.zipCode(args.zipcode, **optionalArgs(args))
    print(data)


async def get_forecast_ll(args: argparse.Namespace) -> None:
    '''Print forecast information for a latitude and longitude'''
    client = WebServiceAPI(args.api_key)
    data = await client.forecast.latLong(*llArgs(args), **optionalArgs(args))
    print(data)


async def get_current_zip(args: argparse.Namespace) -> None:
    '''Print current observation information for a zip code'''
    client = WebServiceAPI(args.api_key)
    data = await client.observations.zipCode(
        args.zipcode,
        **optionalArgs(args),
    )
    print(data)


async def get_current_ll(args: argparse.Namespace) -> None:
    '''Print current observation information for a latitude and longitude'''
    client = WebServiceAPI(args.api_key)
    data = await client.observations.latLong(
        *llArgs(args),
        **optionalArgs(args),
    )
    print(data)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='CLI Client for pyAirNow, which obtains current and '
                    'forecasted air quality information from AirNow. The '
                    'returned JSON data is printed to stdout.',
        epilog='To obtain an API key, visit https://docs.airnowapi.org/login',
    )
    parser.add_argument(
        '-k', '--api-key',
        required=True,
        help='AirNow API Key (required)',
    )

    location_group = parser.add_mutually_exclusive_group(required=True)
    location_group.add_argument(
        '-z', '--zipcode',
        help='Zip Code for Observations or Forecast. May not be used with '
             '--location.',
    )
    location_group.add_argument(
        '-l', '--location',
        help='Latitude and Longitude for Observations or Forecast (must be '
             'formatted as "<latitude>,<longitude>" with an optional space '
             'before the longitude value. May not be used with --zipcode.'
    )

    parser.add_argument(
        '-r', '--distance',
        help='If no reporting area is associated with the zip code or '
             'latitude and longitude, return data from a station within this '
             'distance (in miles)'
    )
    parser.add_argument(
        '-d', '--date',
        help='Desired forecast date. If omitted, the current day forecast is '
             'returned. Format as yyyy-mm-dd'
    )
    parser.add_argument(
        'type', choices=['forecast', 'observations'],
        help='Look up either the forecast or the current observations for the '
             'given location'
    )
    args = parser.parse_args()

    loop = asyncio.get_event_loop()
    if args.type == 'forecast':
        if args.zipcode is not None:
            loop.run_until_complete(get_forecast_zip(args))
        else:
            loop.run_until_complete(get_forecast_ll(args))
    else:
        if args.zipcode is not None:
            loop.run_until_complete(get_current_zip(args))
        else:
            loop.run_until_complete(get_current_ll(args))
